from enum import Enum


class CodeType(Enum):
    OK = 0
    ERROR = 1


def identifyCode(code: int) -> str:
    if code == CodeType.OK.value:
        return "ran normally without errors"
    elif code == CodeType.ERROR.value:
        return "failed due to a fatal error"
    else:
        return "ended with an unknown formatted code, which couldn't be identified."


import os
import os as _os
import os.path as _path
from importlib import import_module as _import_module
from types import ModuleType as _ModuleType

from mavro.parser.lens import LensParser as _LensParser


def public__findService(module: str) -> _ModuleType | None:
    pyname: str = f"{module.removesuffix(".mav")}_requiry.py"
    try:
        if not _path.exists(module):
            raise ImportError(f"Mavro module '{module}' not found in '{_os.getcwd()}'")
        from ..internal.build import build
        build(
            usage="python.exe",
            path=module,
            dist_path=pyname,
            no_delete=True,
            line_loader=_LensParser.stdLoadLinesWithoutEntrypoint,
            run=False
        )
        module_literal: _ModuleType = _import_module(pyname.removesuffix(".py"))
    finally:
        try:
            os.remove(pyname)
        except FileNotFoundError:
            print(f"module error: could not clean {pyname}")
            return
    return module_literal


import os

class DistBuilder:
    def __init__(self, path: str) -> None:
        self.path: str = path
        self.pyname: str = self.path.removesuffix(".mav") + ".py"
    def buildStdDist(self) -> None:
        from ...internal.build import build
        build(
            usage="python.exe",
            path=self.path,
            dist_path=self.pyname,
            no_delete=True,
            run=False
        )
        os.system(f"pyinstaller --onefile --upx-dir=upx --log-level ERROR --noconfirm --distpath . {self.pyname}")
        os.remove(self.pyname)
        os.remove(self.path.removesuffix(".mav") + ".spec")


import sys


def astral_cli() -> None:
    try:
        path: str = sys.argv[1] if len(sys.argv) > 1 else "main.mav"
        print("Hang tight! We're generating your executable as we speak.")
        print("Did you know Mavro executables are compressed with UPX for minimum size and maximum performance.")
        print("Learn more about UPX: https://upx.github.io/")
        from . import DistBuilder
        builder: DistBuilder = DistBuilder(
            path=path
        )
        builder.buildStdDist()
    except KeyboardInterrupt:
        print("Cancelled build operation")
        return
    print("Build operation successful!")

if __name__ == "__main__":
    astral_cli()


def public__style(content: str) -> str:
    return str(content)\
    .replace("$red$", "\033[31m")\
    .replace("$green$", "\033[32m")\
    .replace("$yellow$", "\033[33m")\
    .replace("$blue$", "\033[34m")\
    .replace("$magenta$", "\033[35m")\
    .replace("$cyan$", "\033[36m")\
    .replace("$white$", "\033[37m")\
    .replace("$bold$", "\033[1m")\
    .replace("$reset$", "\033[0m")\
    + "\033[0m"
def public__paint(content: str, style: str) -> str:
    return public__style(f"${style}${content}")
def public__print(content: str) -> None:
    print(public__style(content))


import sys
from typing import Any
from uuid import uuid4
from types import SimpleNamespace


class LineParserResult:
    def __init__(self, output: SimpleNamespace, code: int, error: Exception | None = None) -> None:
        self.output: SimpleNamespace = output
        self.code: int = code
        self.error: Exception | None = error


class LineParserNextSuggestion:
    def __init__(self, suggestion: str) -> None:
        self.suggestion: str = suggestion

    def apply(self, parser: "LineParser") -> "LineParser":
        return parser.next(self.suggestion)


class LineParser:
    def __init__(self, cont: str | None = None) -> None:
        self.id: str = str(uuid4())
        self.cont: str = self.id if cont is None else cont
        self.stack: list[LineParser] = []

    def hasCont(self) -> bool:
        return self.cont != self.id

    def next(self, cont: str) -> "LineParser":
        lp: LineParser = LineParser(cont)
        self.stack.append(lp)
        return lp

    def applyIndentation(self, indent: int) -> "LineParser":
        self.cont = " " * indent + self.cont.replace("\t", "    ").lstrip(" ")
        return self

    def parse(self) -> LineParserResult:
        self.cont = self.cont.rstrip(";")

        def error(exc: Exception) -> LineParserResult:
            return LineParserResult(
                output=SimpleNamespace(),
                code=1,
                error=exc
            )

        output: dict[str, Any] = {
            "cont": "",
            "indent": 0,
            "suggestions": []
        }
        try:
            if not self.hasCont():
                raise TypeError("Cannot parse using a baseline parser.")
            in_string: str = ""
            skip_next: int = 0
            for index, char in enumerate(self.cont):
                if skip_next > 0:
                    skip_next -= 1
                    continue
                try:
                    nearest_char_left: str = self.cont[index - 1]
                except IndexError:
                    nearest_char_left: str = ""
                try:
                    nearest_char_right: str = self.cont[index + 1]
                except IndexError:
                    nearest_char_right: str = ""
                ############################################################### CUSTOM FUNCTIONALITY
                # SERVICE GET FUNCTIONALITY
                if char == ":" and not in_string and (nearest_char_left.isalnum() or nearest_char_left in ")]") and nearest_char_right == ":" and index != 0:
                    output["cont"] += ".public__"
                    skip_next += 1
                elif char == "#" and not in_string:
                    break
                elif char in "\"'":
                    if (char == in_string) and nearest_char_left != "\\":
                        in_string = ""
                        output["cont"] += "\")"
                    elif not in_string:
                        in_string = char
                        output["cont"] += "System.String(f\""
                    else:
                        output["cont"] += char
                else:
                    output["cont"] += char
        except Exception as exception:  # NOQA
            if "--verbose" in sys.argv:
                raise exception
            return error(exception)
        else:
            return LineParserResult(output=SimpleNamespace(**output), code=0)



import os
import sys
from typing import Callable

from ..parser.lens import LensParser, LensParserResult
from ..parser.packaging import Package, PackageImportType


def strap(parser: LensParser, print: Callable = print, *packages: Package, return_dependencies: bool = False) -> str | tuple[str, list[str]]: # NOQA
    parser.loadPackage(
        Package("std", "mavro/pkg"),
        PackageImportType.WILDCARD,
        ""
    )
    for package in packages:
        parser.loadPackage(package, PackageImportType.STD, "")
    result: LensParserResult = parser.parse()
    for error in result.line_errors:
        print(f"internal error (minor): \033[31m{error}\033[0m")
        if "--strict-verbose":
            raise error
    if result.error:
        print(f"internal error (fatal): \033[31m{result.error}\033[0m")
        try:
            input("  press return to submit an issue, or ctrl-c to skip this.")
            os.system("start https://github.com/astridot/mavro/issues/new")
        except KeyboardInterrupt:
            ...
        if "--verbose" in sys.argv:
            raise result.error
    cont: str = result.output.cont if not result.error else ""
    if return_dependencies:
        return cont, result.dependencies
    return cont


class Mesh2D:
    def __init__(self, max_xy: tuple[int, int] | None = None, background: str | None = None) -> None:
        self.max_x, self.max_y = max_xy or 1, 1
        self.background: str = background or "."
        self.tiles: list[tuple[int, int] | str] = [
            (0, 0), self.background,
            (0, 1), self.background,
            (1, 0), self.background,
            (1, 1), self.background,
        ]
    def __mul__(self, value: int) -> None:
        self.max_x, self.max_y = self.max_x * value, self.max_y * value
    def __matmul__(self, value: tuple[tuple[int, int], str]):
        if value[0] in self.tiles:
            self.tiles.pop(self.tiles.index(value[0]) + 1)
            self.tiles.remove(value[0])
        self.tiles.insert(0, value[0])
        self.tiles.insert(1, value[1])
    def public__render(self) -> str:
        val: str = ""
        for y in range(0, self.max_y):
            for x in range(self.max_x):
                try:
                    val += self.tiles[self.tiles.index((x, y)) + 1]
                except (IndexError, ValueError):
                    val += self.background
            val += "\n"
        return val


from typing import Any as _Any

from .std import System as _System


class Base(_System.BaseClass):
    ORIGIN: _Any = None
    def __init__(self) -> None:
        super().__init__()
    def __int__(self) -> int:
        return self.to(int)
    def __str__(self) -> str:
        return self.to(str)
    def to(self, type_) -> _Any:
        return type_(self.ORIGIN)


import os
from contextlib import contextmanager

from ..parser.lens import LensParser
from ..parser.packaging import Package
from .strap import strap


@contextmanager
def create_temporary_file_from_lens(lens: LensParser, *packages: Package, dist_path: str | None = None, no_delete: bool = False) -> None:
    fn: str = dist_path or f"{lens.id}.lens"
    with open(fn, "w") as file:
        file.write(strap(lens, print, *packages))
    yield fn
    if not no_delete:
        os.remove(fn)


from types import ModuleType as requirement # NOQA
from typing import Any as any # NOQA
from typing import Any as unknown # NOQA
from typing import Callable as callable # NOQA
import sys as _sys


true: True = True
false: False = False
null: None = None
iife: callable = lambda _: _()
static: callable = staticmethod

# SYSTEM AND IT'S SUBPROCESSES


class NotStartableError(Exception):
    ...
class MergeOverlapError(Exception):
    ...
class DeprecationError(Exception):
    ...
class EntrypointError(Exception):
    ...
class _SecretFailValue:
    ...

class System:
    PIPE: int = 0
    SPARE: int = 1
    ORIGIN: int = 2
    def public__exit(self, code: int = 0, silent: bool = False) -> None: # NOQA
        if not silent:
            from ..parser.coding import identifyCode
            print(f"\n{identifyCode(code)} (code {code})")
        _sys.exit(code)
    def public__createError(self, name: str) -> type: # NOQA
        return type(name, (Exception,), {})
    def public__importPython(self, module: str) -> "ModuleType": # NOQA
        import importlib
        return importlib.import_module(module)
    def public__ensure(self, # NOQA
                       action: callable,
                       default: any = None,
                       error: type[Exception] = Exception,
                       message: str | None = None
    ):
        try:
            return action()
        except error as exc:
            if str(exc) == message or message is None:
                return default
    def public__redirect(self,
                             action: callable,
                             new: Exception,
                             error: type[Exception] = Exception,
                             message: str | None = None
    ):
        if self.public__ensure(action, _SecretFailValue, error, message) == _SecretFailValue:
            raise new
    def public__merge(self, obj, /, method = None) -> None:
        try:
            name: str = obj.__name__
        except AttributeError:
            name: str = type(obj).__name__
        if method == self.ORIGIN:
            setattr(self, name, obj)
        else:
            for name, value in list(obj.__dict__.items()):
                if hasattr(self, name):
                    if method == self.PIPE:
                        setattr(self, name, value)
                    elif method == System.SPARE:
                        continue
                    else:
                        raise MergeOverlapError(f"{type(self).__name__} and {name} both have the attribute: '{name}'."
                                                "You can specify the 'method=System.PIPE' parameter to overwrite the '{name}'"
                                                "attribute (and any further overlapping attributes. Alternatively, you can specify"
                                                "'method=System.SPARE' to leave overlapping attributes in their original form without changing them."
                        )
                else:
                    setattr(self, name, value)
    class BaseClass:
        def __init__(self) -> None:
            self.__name__ = type(self).__name__
        def __starter__(self, *_, **__) -> None:
            raise NotStartableError(
                f"{type(self).__name__.removeprefix("public__")} doesn't define a starter method, and therefore cannot be started.")
        def __repr__(self) -> str:
            name: str = type(self).__name__
            return f"{"public" if name.startswith("public__") else "local"} class {name.removeprefix("public__")} ({"startable" if hasattr(self, "__starter__") else "not startable"})"


import os
import subprocess
import sys
import pathlib
from typing import Callable

from ..pkg.std import System
from ..parser.lens import LensParser
from ..parser.lines import LineParser
from ..internal.driver import create_temporary_file_from_lens

def build(
        usage: str,
        path: str,
        dist_path: str | None = None,
        no_delete: bool = False,
        line_loader: Callable | None = None,
        run: bool = True
    ) -> str:
    with open(path) as file_literal:
        cont: str = file_literal.read()
    lens: LensParser = LensParser(
        cont=cont,
        baseline=LineParser(),
        line_loader=line_loader
    )
    no_delete = no_delete or "--no-delete" in sys.argv
    with create_temporary_file_from_lens(lens, dist_path=dist_path, no_delete=no_delete) as fn:
        from ..pkg.std import System
        if run:
            result: subprocess.CompletedProcess = System().public__ensure(lambda: subprocess.run(
                f"{usage} {fn}",
                stderr=subprocess.PIPE,
                text=True,
                shell=True
            ), error=KeyboardInterrupt) # NOQA
            if result:
                if result.stderr:
                    if "--verbose" in sys.argv:
                        print(result.stderr)
                    else:
                        try: info: str = result.stderr.split("\n")[-2].split(":", 1)[1].lstrip()
                        except Exception: # NOQA
                            info: str = "this error was passed with no message nor traceback."
                        try: name: str = result.stderr.split("\n")[-2].split(":", 1)[0]
                        except Exception: # NOQA
                            name: str = "ErrorNotFound"
                        if info == "'System' object has no attribute 'BaseString'":
                            name = "StringError"
                            info = "Attempted usage of String object without it's dedicated System merge."
                        print(f"fatal {name}: \033[31m{info}\033[0m")
    return fn
def build_from_sys_argv() -> None:
    args: list[str] = sys.argv
    try: args[1]
    except IndexError:
        args.append(".")
    usage: str = "python.exe"
    if "-u" in args:
        try:
            usage = args[args.index("-u") + 1]
        except IndexError:
            print("'-u' argument expected one parameter.")
            sys.exit(1)
    if args[1] in ["--version", "-v"]:
        from .. import version
        print(f"Mavro version: {version}")
        sys.exit(0)
    elif args[1] == "--github":
        os.system("start https://github.com/astridot/mavro")
        sys.exit(0)
    elif args[1] == "--pypi":
        os.system("start https://pypi.org/project/mavro")
        sys.exit(0)
    elif args[1] == "--author":
        os.system("start https://github.com/astridot")
        sys.exit(0)
    elif args[1].startswith("-"):
        args.append(args[1])
        args[1] = "."
    if os.path.exists(args[1]):
        if os.path.isdir(args[1]):
            args[1] = str(pathlib.Path(args[1]) / "main.mav")
        if "-c" in args or "--create" in args:
            with open(args[1], "w") as file:
                file.write("entrypoint\n  .print 'main.mav file was created!'\nend")
        if not os.path.exists(args[1]):
            print(f"'{args[1]}' does not exist, and cannot be executed.")
            return
        build(
            usage=usage,
            path=args[1]
        )
    else:
        print(f"'{args[1]}' does not exist, and cannot be executed.")
if __name__ == "__main__":
    build_from_sys_argv()



from typing import Iterator

from .std import System as _System
from .remote import Base as _Base


class String(str):
    def equals(self, other: str) -> bool:
        return self == other

class BaseString(_Base, String):
    ORIGIN: str = ""
    def toStr(self) -> str:
        return self.to(str)


import sys
from types import SimpleNamespace
from typing import Any, Callable
from uuid import uuid4

from .coding import identifyCode
from .lines import LineParser, LineParserResult
from .packaging import Package, PackageImportType


class LensParserResult(LineParserResult):
    def __init__(self, output: SimpleNamespace, code: int, error: Exception | None = None, line_errors: list[Exception] | None = None, dependencies: list[str] | None = None) -> None:
        super().__init__(output, code, error)
        self.line_errors: list[Exception] = line_errors or []
        self.dependencies: list[str] = dependencies or []


class LensParser:
    LINE_LOADER_BEFORE: list[str] = """
from mavro.pkg.std import *
const System System = System()
package console
package string
System::merge console.Console, method=System.ORIGIN
System::merge string, method=System.SPARE
startprocess System.Console
const console.Console Console = System.Console
const string.String String = System.String
savelocation

function _deprecated name, sub=null
    function wrapper *_, **__
        raise DeprecationError("'{name}' is deprecated and cannot be used. {'Use \\'{}\\' instead.'.format(sub) if sub else ''}")
    end
    return wrapper
end
const callable print = _deprecated("print", "Console::print")
const callable input = _deprecated("input", "Console::input")
const callable exit = _deprecated("exit", "System::exit")
const callable str = _deprecated("str", "System.String")
del console, string
""".split("\n")
    LINE_LOADER_AFTER: list[str] = """
try
    remark __entrypoint__
catch NameError
else
    only private
        __entrypoint__
        System::exit 0
    end
end
""".split("\n")
    def __init__(self, cont: str, baseline: LineParser, line_loader: Callable | None = None) -> None:
        """The line_loader parameter shouldn't need to be changed."""
        self.cont: str = cont
        self.id: str = str(uuid4())
        self.lns: list[str | LineParserResult] = (line_loader or self.stdLoadLines)(self.cont)
        self.baseline: LineParser = baseline
    @staticmethod
    def stdLoadLines(cont: str) -> list[str]:
        lns: list[str] = [*LensParser.LINE_LOADER_BEFORE,
                          *cont.split("\n"),
                          *LensParser.LINE_LOADER_AFTER
        ]
        return lns
    @staticmethod
    def stdLoadLinesWithoutEntrypoint(cont: str) -> list[str]:
        lns: list[str] = [*LensParser.LINE_LOADER_BEFORE,
                          *cont.split("\n")
        ]
        return lns
    @staticmethod
    def loadPackage(package: Package, import_type: PackageImportType, arg: str) -> str | Exception:
        if package.origin != "mavro/pkg":
            return ImportError("Tried retrieving a pkg that doesn't seem to originate from mavro's verified pkg source (mavro/pkg)")
        text: str | Exception = package.getImportStatement(import_type, arg)
        return text
    @staticmethod
    def _includeKwCheck(cont: str) -> bool:
        import keyword
        kws: list[str] = list(keyword.kwlist)

        for kw in kws:
            if cont.startswith(f"{kw} ") or cont == kw:
                return True
        return False
    def parse(self) -> LensParserResult:
        def error(exc: Exception) -> LensParserResult:
            return LensParserResult(
                output=SimpleNamespace(),
                code=1,
                error=exc,
                line_errors=line_errors
            )
        output: dict[str, Any] = {
            "cont": ""
        }
        indent: int = 0
        line_errors: list[Exception] = []
        dependencies: list[str] = []
        try:
            for ln_num, ln in enumerate(self.lns, start=1):
                if isinstance(ln, str):
                    parser: LineParser = self.baseline.next(ln)
                elif isinstance(ln, LineParser):
                    parser: LineParser = ln
                else:
                    raise TypeError(f"Invalid line type on stack n. {ln_num} most likely inserted due to a faulty suggestion by a line parser.\n"
                                    f"Expected type str or LineParserResult, got {type(ln).__name__}.")
                parser.applyIndentation(indent)
                result: LineParserResult = parser.parse()
                original_indent: int = indent
                if "--verbose" in sys.argv:
                    print(f"Parsing of stack n. {"0" * (3 - len(str(ln_num)))}{ln_num} indented {"0" * (2 - len(str(int(indent / 4))))}{int(indent / 4)} layers {identifyCode(result.code)} (returned code {result.code}) ~ {result.output.cont}")
                if result.error:
                    line_errors.append(result.error)
                    continue
                for suggestion in result.output.suggestions:
                    self.lns.insert(ln_num + 1, suggestion.apply(parser))
                if isinstance(result.output, SimpleNamespace):
                    cont: str = result.output.cont.strip().removeprefix("local ")
                else:
                    print(f"Unexpected type for result.output: {type(result.output)}")
                    continue
                if cont.startswith("let ") or cont.startswith("const "):
                    parts: list[str] = cont.split(" ", 2)
                    try:
                        cont = f"{parts[2][:parts[2].index("=")]}: {parts[1]}{parts[2][parts[2].index("="):]}"
                        if parts[1] == "str":
                            cont = f"str()"
                    except IndexError:
                        raise TypeError(f"Not enough parameters for variable definition. Usage: {parts[0]} <type> <name> = <value>`")
                elif cont.startswith("remark "):
                    cont = cont.removeprefix("remark ")
                elif cont.startswith("import "):
                    parts: list[str] = cont.split(" ", 4)
                    if "import mavro.pkg.requiry as requiry" in output["cont"] and "." not in parts[1]:
                        alias: str = ""
                        if len(parts) > 2:
                            if parts[2] == "as":
                                alias = parts[3]
                        cont = f"{alias or parts[1]} = System.public__ensure(lambda: System.public__importPython('{parts[1]}'), None, ModuleNotFoundError) or requiry.public__findService('{parts[1]}.mav')"
                    dependencies.append(parts[1])
                elif cont.startswith("from "):
                    parts: list[str] = cont.split(" ", 6)
                    if parts[2] != "import":
                        raise SyntaxError("'from' keyword expected 'import'")
                    if "import mavro.pkg.requiry as requiry" in output["cont"] and "." not in parts[1]:
                        alias: str = ""
                        if len(parts) > 4:
                            if parts[4] == "as":
                                alias = parts[5]
                        cont = f"{alias or parts[3]} = (System.public__ensure(lambda: System.public__importPython('{parts[1]}'), None, ModuleNotFoundError) or requiry.public__findService('{parts[1]}.mav')).{parts[3]}"
                    dependencies.append(parts[1])
                elif cont.startswith("public const "):
                    parts: list[str] = cont.split(" ", 4)
                    try:
                        cont = f"public__{parts[3][:parts[3].index("=")]}: {parts[2]}{parts[3][parts[3].index("="):]}"
                        if parts[2] == "str":
                            cont = f"str()"
                    except IndexError:
                        raise TypeError(f"Not enough parameters for public variable definition. Usage: `public {parts[1]} <type> <name> = <value>`")
                elif cont.startswith("public let "):
                    raise TypeError("Variables declared with 'let' cannot be public.")
                elif cont.startswith("function "):
                    parts: list[str] = cont.split(" ", 2)
                    cont = f"def {parts[1]}({parts[2] if len(parts) > 2 else ""}):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("method "):
                    parts: list[str] = cont.split(" ", 2)
                    cont = f"def {parts[1]}(self, {parts[2] if len(parts) > 2 else ""}):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("apply "):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"@{parts[1]}"
                elif cont.startswith("remote "):
                    parts: list[str] = cont.split(" ", 3)
                    cont = f"@lambda _: _()\n{" " * original_indent}class {parts[2]}({parts[1]}):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("root "):
                    parts: list[str] = cont.split(" ", 2)
                    if parts[1] == "...":
                        cont = ""
                    else:
                        cont = f"@{parts[1]}"
                    self.lns.insert(ln_num, parts[2])
                elif cont.startswith("special method "):
                    parts: list[str] = cont.split(" ", 3)
                    cont = f"def {parts[2]}(self, {parts[3] if len(parts) > 3 else ""}):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("public function "):
                    parts: list[str] = cont.split(" ", 3)
                    cont = f"def public__{parts[2]}({parts[3] if len(parts) > 3 else ""}):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("public method "):
                    parts: list[str] = cont.split(" ", 3)
                    cont = f"def public__{parts[2]}(self, {parts[3] if len(parts) > 3 else ""}):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("require "):
                    parts: list[str] = cont.split(" ", 2)
                    if "import mavro.pkg.requiry as requiry" not in output["cont"]:
                        raise SyntaxError("Attempted fetching of Mavro module while the requiry package wasn't imported.")
                    cont = f"{parts[1]} = requiry.public__findService('{parts[1]}.mav')"
                elif cont.startswith("def "):
                    raise SyntaxError("Python 'def' keyword is not supported in Mavro. Use 'function' instead")
                elif cont.startswith("if "):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"if {parts[1]}:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("else if "):
                    original_indent -= 4
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"elif {parts[1]}:\n{" " * (original_indent + 4)}..."
                elif cont.startswith("elif "):
                    raise SyntaxError("Python 'elif' keyword is not supported in Mavro. Use 'else if' instead")
                elif cont == "else":
                    original_indent -= 4
                    cont = f"else:\n{" " * (original_indent + 4)}..."
                elif cont == "try":
                    cont = f"try:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont == "finally":
                    original_indent -= 4
                    cont = f"finally:\n{" " * (original_indent + 4)}..."
                elif cont == "entrypoint":
                    cont = f"def __entrypoint__() -> int:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont == "end":
                    cont = f"# end"
                    indent -= 4
                elif cont.startswith("end "):
                    self.lns.insert(ln_num + 1, cont.removeprefix("end "))
                    indent -= 4
                elif cont == "only private":
                    cont = "if __name__ == \"__main__\":"
                    indent += 4
                elif cont == "only public":
                    cont = "if __name__ != \"__main__\":\n"
                    indent += 4
                elif cont == "savelocation":
                    cont = f"from types import SimpleNamespace\n{" " * original_indent}here = SimpleNamespace(**(globals() | locals()))\n{" " * original_indent}del SimpleNamespace"
                elif cont.startswith("catch "):
                    original_indent -= 4
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"except {parts[1]}:\n{" " * (original_indent + 4)}..."
                elif cont.startswith("except "):
                    raise SyntaxError("Python 'except' keyword is not supported in Mavro. Use 'catch' instead")
                elif cont.startswith("while "):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"while {parts[1]}:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("for "):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"for {parts[1]}:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("constructor"):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"def __init__(self, {parts[1] if len(parts) > 1 else ""}):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("extends constructor"):
                    parts: list[str] = cont.split(" ", 2)
                    cont = f"def __init__(self, {parts[2] if len(parts) > 2 else ""}):\n{" " * (original_indent + 4)}super().__init__({parts[2] if len(parts) > 2 else ""})"
                    indent += 4
                elif cont.startswith("starter"):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"def __starter__(self, {parts[1] if len(parts) > 1 else ""
                    }):\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("startprocess "):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"{parts[1]}.__starter__({parts[2] if len(parts) > 2 else ""})"
                elif cont.startswith("until "):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"while not {parts[1]}:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("class "):
                    parts: list[str] = cont.split(" ", 3)
                    if parts[2] == "extends" and len(parts) > 3:
                        cont = f"class {parts[1]}({parts[3]}):\n{" " * (original_indent + 4)}..."
                    else:
                        cont = f"class {parts[1]}:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("public class "):
                    parts: list[str] = cont.split(" ", 4)
                    if parts[3] == "extends" and len(parts) > 3:
                        cont = f"class public__{parts[2]}({parts[4]}):\n{" " * (original_indent + 4)}..."
                    else:
                        cont = f"class public__{parts[2]}:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("manager "):
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"with {parts[1]}:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("with "):
                    raise SyntaxError("Python 'with' keyword is not supported in Mavro. Use 'manager' instead")
                elif cont.startswith("openfile "):
                    parts: list[str] = cont.split(" ", 2)
                    cont = f"with open({parts[1]}, {parts[2]}) as file:\n{" " * (original_indent + 4)}..."
                    indent += 4
                elif cont.startswith("package "):
                    parts: list[str] = cont.split(" ", 1)
                    try:
                        package_name: str = parts[1]
                    except IndexError:
                        raise ImportError("Not enough arguments for package import.")
                    try:
                        package: Package = Package(
                            name=package_name,
                            origin="mavro/pkg"
                        )
                        package_result: str | Exception = self.loadPackage(package, PackageImportType.AS, package.name)
                        if isinstance(package_result, Exception):
                            raise package_result
                        cont = package_result
                    except Exception as exception:
                        raise exception
                elif self._includeKwCheck(cont):
                    ...
                else:
                    parts: list[str] = cont.split(" ", 1)
                    cont = f"{parts[0]}({parts[1] if len(parts) > 1 else ""})"
                cont = f"{" " * original_indent}{cont}"
                output["cont"] += cont + "\n" # NOQA
                indent += result.output.indent # NOQA
        except Exception as exception:  # NOQA
            if "--verbose" in sys.argv:
                raise exception
            return error(exception)
        else:
            return LensParserResult(output=SimpleNamespace(**output), code=0, line_errors=line_errors)


from .std import System as _System
class Console(_System.BaseClass):
    def __init__(self) -> None:
        self.__activated: bool = False
    def __assertActivated(self, fn: str | None = None) -> None:
        if not self.__activated:
            raise BlockingIOError(f"Could not execute 'Console::{fn or "unknownService"}' because the Console service is not activated.")
    def __starter__(self) -> None:
        self.__activated = True
    def public__print(self, obj: any) -> None:
        self.__assertActivated("print")
        print(obj)
    def public__input(self, obj: any) -> str:
        self.__assertActivated("input")
        print(obj, end="")
        return input()

Console = Console()


from setuptools import setup, find_packages

def read_long_description():
    with open('README.md', encoding='utf-8') as f:
        return f.read()

from mavro import version

setup(
    name='Mavro',
    version=version,
    author='astridot',
    author_email='pixilreal@gmail.com',
    description='An easy to understand language built from Python.',
    long_description=read_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
    install_requires=["pyinstaller"],
    entry_points={
        'console_scripts': [
            'mavro=mavro.internal.build:build_from_sys_argv'
        ]
    },
)


"""
Astral provides the 'mavro/distbuilder/astral' directory that Mavro uses to build files into executable format.
Astral is a .py->.exe distribution builder that uses the PyInstaller and UPX. Astral, not including PyInstaller or UPX,
is licensed by the MavroLang development foundation. Astral may only be installed for free as part of MavroLang,
or purchased separately with a separate license from Astral distributors.

PyInstaller license: https://pyinstaller.org/en/stable/license.html
UPX license: https://github.com/upx/upx/blob/devel/LICENSE
"""


from enum import Enum, auto
from importlib import import_module
from types import ModuleType
from typing import Any


class PackageImportType(Enum):
    WILDCARD = auto()
    FROM = auto()
    STD = auto()
    AS = auto()


class Package:
    def __init__(self, name: str, origin: str | None = None) -> None:
        self.name: str = name
        self.origin: str | None = origin
    def __repr__(self) -> str:
        return f"{self.origin.replace("/", ".").replace("\\", ".")}.{self.name}"
    def getImportStatement(self, import_type: PackageImportType, arg: str) -> str:
        match import_type:
            case PackageImportType.WILDCARD:
                return f"from {repr(self)} import *"
            case PackageImportType.FROM:
                return f"from {repr(self)} import {arg}"
            case PackageImportType.STD:
                return f"import {repr(self)}"
            case PackageImportType.AS:
                return f"import {repr(self)} as {arg}"
            case _:
                raise ValueError(f"Unsupported or incorrect import type: {import_type}")
    def getModule(self) -> ModuleType | Exception:
        try: return import_module(repr(self), __package__)
        except ModuleNotFoundError: return ImportError(f"Package '{self.name}' from origin '{self.origin}' does not exist.")
    def getModuleItem(self, item: str) -> Any | Exception:
        module: ModuleType | Exception = self.getModule()
        if isinstance(module, Exception):
            raise module
        if hasattr(module, item):
            return getattr(module, item)
        else:
            return AttributeError(f"Package '{self.name}' from origin '{self.origin}' does not have an attribute with the name '{item}'.")